import copy
import warnings
from collections import defaultdict

__all__ = [
    "get_all_keys",
    "get_occurrences_and_values",
    "get_occurrence_of_value",
    "get_occurrence_of_key",
    "nested_update",
    "nested_delete",
    "nested_alter",
    "nested_lookup",
]


# TODO: get rid of this global values_list
values_list = []


def nested_lookup(key, document, wild=False, with_keys=False):
    """Lookup a key in a nested document, return a list of values"""
    if with_keys:
        d = defaultdict(list)
        for k, v in _nested_lookup(key, document, wild=wild, with_keys=with_keys):
            d[k].append(v)
        return d
    return list(_nested_lookup(key, document, wild=wild, with_keys=with_keys))


def _is_case_insensitive_substring(a, b):
    """return True if `a` is a case insensitive substring of `b`, else False"""
    return str(a).lower() in str(b).lower()


def _nested_lookup(key, document, wild=False, with_keys=False):
    """Lookup a key in a nested document, yield a value"""
    if isinstance(document, list):
        for d in document:
            for result in _nested_lookup(key, d, wild=wild, with_keys=with_keys):
                yield result

    if isinstance(document, dict):
        for k, v in document.items():
            if key == k or (wild and _is_case_insensitive_substring(key, k)):
                if with_keys:
                    yield k, v
                else:
                    yield v
            if isinstance(v, dict):
                for result in _nested_lookup(key, v, wild=wild, with_keys=with_keys):
                    yield result
            elif isinstance(v, list):
                for d in v:
                    for result in _nested_lookup(
                        key, d, wild=wild, with_keys=with_keys
                    ):
                        yield result


def get_all_keys(dictionary):
    """
    Method to get all keys from a nested dictionary as a List
    Args:
        dictionary: Nested dictionary
    Returns:
        List of keys in the dictionary
    """
    result_list = []

    def recrusion(document):
        if isinstance(document, list):
            for list_items in document:
                recrusion(document=list_items)
        elif isinstance(document, dict):
            for key, value in document.items():
                result_list.append(key)
                recrusion(document=value)
        return

    recrusion(document=dictionary)
    return result_list


def get_occurrence_of_key(dictionary, key):
    """
    Method to get occurrence of a key in a nested dictionary

    Args:
        dictionary: Nested dictionary
        key: Key to search for the occurrences
    Return:
        Number of occurrence (Integer)
    """
    return _get_occurrence(dictionary=dictionary, item="key", keyword=key)


def get_occurrences_and_values(items, value):
    """
    Method to get occurrence of a value in a nested list of dictionary

    Args:
        items: list of dictionary: Nested dictionary
        value: Value to search for the occurrences

    Return:
        Dict where the key is the value arg and his value is a new
        dict with occurrences and values
    """
    occurrences = {}
    occurrence = 0
    value_list = []

    for item in items:
        occurrence_result, values = _get_occurrence_with_values(
            dictionary=item, item="value", keyword=value
        )
        occurrence = occurrence + occurrence_result
        if occurrence_result:
            value_list.extend(values)

    occurrences[value] = {"occurrences": occurrence, "values": value_list}

    return occurrences


def _get_occurrence_with_values(dictionary, item, keyword):
    occurrence = [0]

    result_recursion = _recursion(dictionary, item, keyword, occurrence, True)

    global values_list
    values_list = []

    return occurrence[0], result_recursion


def get_occurrence_of_value(dictionary, value):
    """
    Method to get occurrence of a value in a nested dictionary

    Args:
        dictionary: Nested dictionary
        value: Value to search for the occurrences
    Return:
        Number of occurrence (Integer)
    """
    return _get_occurrence(dictionary=dictionary, item="value", keyword=value)


def _recursion(dictionary, item, keyword, occurrence, with_values=False):
    global values_list

    if item == "key":
        if dictionary.get(keyword) is not None:
            occurrence[0] += 1
    elif keyword in list(dictionary.values()):
        occurrence[0] += list(dictionary.values()).count(keyword)
        if with_values:
            values_list.append(dictionary)
    for key, value in dictionary.items():
        if isinstance(value, dict):
            _recursion(value, item, keyword, occurrence, with_values)
        elif isinstance(value, list):
            for list_items in value:
                if hasattr(list_items, "items"):
                    _recursion(list_items, item, keyword, occurrence, with_values)
                elif list_items == keyword:
                    occurrence[0] += 1 if item == "value" else 0

    if values_list:
        return values_list


def _get_occurrence(dictionary, item, keyword):
    """
    Method to get occurrence of a key or value in a nested dictionary

    Args:
        dictionary: Nested dictionary
        item: Mostly (key or value)
        keyword: key word to find occurrence
    Return:
        Number of occurrence of the given keyword in the dict
    """
    occurrence = [0]
    _recursion(dictionary, item, keyword, occurrence)

    global values_list
    values_list = []

    return occurrence[0]


def nested_delete(document, key, in_place=False):
    if not in_place:
        document = copy.deepcopy(document)
    return _nested_delete(document=document, key=key)


def _nested_delete(document, key):
    """
    Method to delete a key->value pair from a nested document
    Args:
        document: Might be List of Dicts (or) Dict of Lists (or)
         Dict of List of Dicts etc...
        key: Key to delete
    Return:
        Returns a document that includes everything but the given key
    """
    if isinstance(document, list):
        for list_items in document:
            _nested_delete(document=list_items, key=key)
    elif isinstance(document, dict):
        if document.get(key):
            del document[key]
        for dict_key, dict_value in document.items():
            _nested_delete(document=dict_value, key=key)
    return document


def nested_update(document, key, value, in_place=False, treat_as_element=True):
    """
    Method to update a key->value pair in a nested document

    Args:
        document: Might be List of Dicts (or) Dict of Lists (or)
            Dict of List of Dicts etc...

        key: Key to update the value

        value: Value to set

        in_place (bool):
            True: modify the dict in place;
            False: create a deep copy of the dict and modify it
            Defaults to False

        treat_as_element (bool):
            True: if a list is provided as "value", the function trys
                to match the list elements to the occurences of the key.
                If the key occures more often than the provided list has
                elements, the first element gets recycled.
            False: the provided list is treated as one scalar value and
                will be set as value to every key that matches.
            Defaults to True (because of backwards portability of the package).

    Return:
        Returns a document that has updated key, value pair.
    """

    # check if a list or scalar value is provided and create a list
    # from the scalar value
    # check the length of the list and provide it to _nested_update
    if not treat_as_element and not isinstance(value, list):
        raise Exception("The value must be a list when treat_as_element is False.")
    elif treat_as_element:
        value = [value]

    if not in_place:
        document = copy.deepcopy(document)
    return _nested_update(document=document, key=key, value=value)


def _nested_update(document, key, value):
    """
    Method to update a key->value pair in a nested document.
    If the number of passed values is less than the number of key matches
    when scanning for updates, use last value as default.
    Args:
        document: Might be List of Dicts (or) Dict of Lists (or)
            Dict of List of Dicts etc...
        key (str): Key to update the value
        value (list): value(s) which should be used for replacement purpouse
    Return:
        Returns a document that has updated key, value pair.
    """
    if isinstance(document, list):
        for list_items in document:
            _nested_update(document=list_items, key=key, value=value)
    elif isinstance(document, dict):
        for dict_key, dict_value in document.items():
            if dict_key == key:
                document[key] = value[0]
                if len(value) > 1:
                    value.pop(0)
            _nested_update(document=dict_value, key=key, value=value)
    return document


def nested_alter(
    document,
    key,
    callback_function=None,
    function_parameters=None,
    conversion_function=None,
    wild_alter=False,
    in_place=False,
):
    """
    Method to alter all values of the occurences of the key "key".
    The provided callback_function is used to alter the scalar values
    Args:
        document: Might be List of Dicts (or) Dict of Lists (or)
            Dict of List of Dicts etc...
        key: Key to update the value
        callback_function :A callback function which alters a scalar value
            HINT: You should be aware that not every element might be of
            the same type, please check this in your function!
        function_parameters (list):
            If the callback_function has additional input arguments except
            the scalar value, please specify those in this list.
        conversion_function: A conversion function like str() which should be
            applied to every found value before it is passed to the
            "callback_function"
        wild_alter: Find matching elements via wild-match by the given keys
            and alter those.
        HINT: Keep in mind that the wild-match might return unexpected types!
        in_place (bool):
            True: modify the dict in place;
            False: create a deep copy of the dict and modify it
            Defaults to False
    Return:
        Returns a document that has updated key, value pair.
    """
    # check if a list or scalar value is provided and create a list from
    # the scalar value
    # check the length of the list and provide it to _nested_update
    if isinstance(key, list):
        key_len = len(key)
    else:
        key = [key]
        key_len = len(key)

    if not in_place:
        document = copy.deepcopy(document)
    return _nested_alter(
        document=document,
        keys=key,
        callback_function=callback_function,
        function_parameters=function_parameters,
        conversion_function=conversion_function,
        wild_alter=wild_alter,
        in_place=in_place,
        key_len=key_len,
    )


def _call_callback(
    value_list, callback_function, function_parameters, conversion_function
):
    """
    internal helper to call the callback function
    """
    return_list = []
    # loop over all values
    for value in value_list:
        # apply the conversion function
        if conversion_function is not None:
            value = conversion_function(value)
        # if functions arguments are present, expand the list to variables
        # via the magic operator *
        if function_parameters:
            trans_val = callback_function(value, *function_parameters)
        else:
            trans_val = callback_function(value)
        # append the transformed element to the list
        return_list.append(trans_val)
    return return_list


def _nested_alter(
    document,
    keys,
    callback_function,
    function_parameters,
    conversion_function,
    wild_alter,
    in_place,
    key_len,
):
    # return data if no callback_function is provided
    if callback_function is None:
        warnings.warn("Please provide a callback_function to nested_alter().")
        return document

    # iterate over all given keys in the list
    for key in keys:
        # try to find the key:
        findings = nested_lookup(key, document, with_keys=True, wild=wild_alter)
        for k, v in findings.items():
            trans_val = _call_callback(
                v, callback_function, function_parameters, conversion_function
            )
            # use the transformed value and apply the update to the key
            # (dont treat the lists as elements here)
            document = nested_update(
                document, k, trans_val, in_place=in_place, treat_as_element=False
            )

    return document



import json
from typing import Type

from .base_parser import BaseParser


class JsonParser(BaseParser):
    def __init__(self, file_contents: str):
        super().__init__(file_contents)

    @property
    def decode_error(self) -> Type[Exception]:
        return json.decoder.JSONDecodeError

    def loads(self, contents: str) -> dict:
        return json.loads(contents)

    def dumps(self, loaded_contents: dict) -> str:
        return json.dumps(loaded_contents, indent=4)



from abc import ABC, abstractmethod
from typing import Any


class AbstractParser(ABC):
    """Abstract Parser is the API contract for a parser for
    use in the ConfigFile object.

    Every parser in ConfigFile implements this parser, and
    any user created parsers must implement this parser in
    order for this package to be able to use it.

    Every key is specified in a dot (.) syntax
    e.g. retrieving.some.deeply.nested.value

    If there is no dot in the key, it is assumed we are
    performing an action on a top level key, which could
    be an entire section.

    Args:
        file_contents: The stringified version of a file.
    """

    @abstractmethod
    def __init__(self, file_contents: str) -> None:
        self.file_contents = file_contents

    @property
    @abstractmethod
    def parsed_content(self) -> dict:
        """To be able to index into the ConfigFile object,
        we have to expose the parsed dictionary object.

        Raises:
            KeyError: If the key we are trying to index into
            does not exist.

        Returns:
            The internal contents parsed as a dictionary.
        """
        raise NotImplementedError

    @abstractmethod
    def reset_internal_contents(self, file_contents: str) -> None:
        """Reset the state of this parser to a new file.

        When restoring to the original file in ConfigFile,
        we need a way to reset the internal contents to this
        new file.

        While this may be as simple as setting the file_contents
        attribute value to the passed in file_contents argument,
        it likely would not be. This is because the initial
        passed in file contents would likely be parsed into some
        special form for the parser to work with more easily.

        Args:
            file_contents: The file contents of the new file.
        """
        raise NotImplementedError

    @abstractmethod
    def get(self, key: str) -> Any:
        """Retrieve the value of a key from the working file.

        Args:
            key: The key to retrieve.

        Returns:
            Whatever value the given key happens to contain.

        Raises:
            KeyError: If the key to retrieve does not exist.
        """
        raise NotImplementedError

    @abstractmethod
    def set(self, key: str, value: Any) -> None:
        """Set the value of a key in the working file.

        If the key does not exist, it should be created.

        Args:
            key: The key to set.
            value: The value to set the key to.
        """
        raise NotImplementedError

    @abstractmethod
    def delete(self, key: str) -> None:
        """Deletes a key from the working file.

        Args:
            key: The key we would like to remove.

        Raises:
            KeyError: If the key to remove does not exist.
        """
        raise NotImplementedError

    @abstractmethod
    def __str__(self) -> str:
        """
        Stringify the working file.

        Since the ConfigFile does not write out to disk
        after every operation, this method is what is
        used as the output for saving the file again.
        This parser should never write out to the disk.

        Returns:
            The file we are working with as a string.
        """
        raise NotImplementedError

    @abstractmethod
    def has(self, key: str, wild: bool = False) -> bool:
        """Check whether the given key is in the parsed working file.

        This should not behave like a wild card, as in check
        if the key exists anywhere in the working file. Instead,
        if a key is given without a dot, it should be assumed we
        are checking the top level.

        Args:
            key: The key to search for.
            wild: Optionally search for any occurrence of key in the file.

        Returns:
            True if the working file has the key. False otherwise.
        """
        raise NotImplementedError



from pathlib import Path
from typing import List, Union


def split_on_dot(line: Union[str, Path], only_last_dot: bool = False) -> List[str]:
    """
    Split a string on the dot character (.).

    Args:
        line: The line ot split on.
        only_last_dot: Only split on the last occurrence of the dot.

    Raises:
        ValueError: if the line does not have a dot.
    """
    if isinstance(line, Path):
        line = str(line)

    if "." not in line:
        raise ValueError(f"The given string does not have a dot to split on: {line}")

    return line.rsplit(".", 1) if only_last_dot else line.split(".")


class Default:
    """
    Default is used for the `default` value in ConfigFile's `get`.

    Previously, the default value for `get` was None. However, then
    the user cannot have `get` return a default value of None. So this
    class is used instead to get around that limitation.
    """

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "Default Value: {} ({})".format(self.value, type(self.value))



from pathlib import Path
from typing import Callable, Tuple, Type, Union

import pytest

from config_file.config_file_path import ConfigFilePath
from config_file.parsers.abstract_parser import AbstractParser


@pytest.fixture
def tmp_file(tmp_path) -> Callable[[str, str], Path]:
    """
    Retrieve the path to a a temporary file.

    Args:
        file_name: The name of the file to create.
        file_contents: The contents of the temporary file.

    Returns:
        A Path pointing at the temporary file that has been created.
    """

    def func(file_name: str, file_contents: str) -> Path:
        file = tmp_path / file_name
        file.touch()
        file.write_text(file_contents, encoding="utf-8")
        return file

    return func


@pytest.fixture
def template_file(tmp_file: Callable[[str, str], Path]) -> Callable[[str, str], Path]:
    """
    Retrieve a path to a file from the test templates/

    All these files mirror each other for every file type
    so that one test can run on all the different file types.

    Args:
        file_type: The filetype to retrieve i.e. "ini"
        template_name: The name of the template to retrieve.
            Defaults to the "default" template. i.e. the template
            with the name "default" with it's respective file extension.

    Returns:
        A Path object point to that test template.
    """
    template_dir = Path(__file__).parent / "templates"

    def func(file_type: str, template_name: str = "default") -> Path:
        file_name = f"{template_name}.{file_type}"
        template = template_dir / file_type / file_name

        return tmp_file(file_name, template.read_text(encoding="utf-8"))

    return func


@pytest.fixture
def template_original_file() -> Callable[[Path], Path]:
    """
    Retrieve the "original" version of a template test file.

    None of the "original" files are actually stored in templates/.
    Instead, it is created on the fly with the same contents. The
    purpose here is that some functionality requires a filename.original.ext
    name, namely, restore_original().

    Args:
        template_path: The path to the temporary template file we're using.

    Returns:
        A Path object pointing to the newly created original template file.
    """

    def func(template_path: Union[str, Path]):
        if isinstance(template_path, str):
            template_path = Path(template_path)

        original_path = ConfigFilePath(template_path).original_path
        original_path.touch()
        original_path.write_text(template_path.read_text(encoding="utf-8"))

        return original_path

    return func


@pytest.fixture
def template_and_parser(
    template_file: Callable[[str, str], Path]
) -> Callable[[str, str], Tuple[Path, Type[AbstractParser]]]:
    """
    Retrieve the parser and test template for a particular
    supported file type.

    Args:
        file_type: The parser type to retrieve i.e. "ini" -> IniParser
        template_name: The template to retrieve. Defaults to the
            "default" template file for each extension.

    Returns:
        The template and parser corresponding to the file_type with
        the template file read in.
    """

    def func(
        file_type: str, template_name: str = "default"
    ) -> Tuple[Path, Type[AbstractParser]]:
        test_file = template_file(file_type, template_name)
        return test_file, ConfigFilePath(test_file).parser

    return func


@pytest.fixture
def templated_parser(template_and_parser) -> Callable[[str, str], Type[AbstractParser]]:
    """
    Retrieve a parser for a particular supported file type with
    a test template already read into it. Small helper around
    template_and_parser that only returns the parser.

    Args:
        file_type: The parser type to retrieve i.e. "ini" -> IniParser
        template_name: The template to retrieve. Defaults to the
            "default" template file for each extension.

    Returns:
        The parser corresponding to the file_type with the template file
        read in.
    """

    def func(file_type: str, template_name: str = "default") -> Type[AbstractParser]:
        return template_and_parser(file_type, template_name)[1]

    return func



from unittest.mock import patch

import pytest

from config_file.parsers.toml_parser import TomlParser


def test_that_import_error_is_raised_if_no_tomlkit():
    """
    config_file.toml_parser.TomlParser

    Ensure an import error is raised if tomlkit is not
    installed.
    """
    with pytest.raises(ImportError) as error:
        with patch("builtins.__import__") as mock_import:
            mock_import.side_effect = ImportError
            TomlParser("")

    assert (
        "Install the `toml` extra first with `pip install config-file[toml]`."
        in str(error)
    )



from typing import Type

from .base_parser import BaseParser


class TomlParser(BaseParser):
    def __init__(self, file_contents: str):
        try:
            super().__init__(file_contents)
        except ImportError:
            raise ImportError(
                "It doesn't appear `tomlkit` is installed, but a toml "
                "file was attempted to be used. Install the `toml` "
                "extra first with `pip install config-file[toml]`."
            )

    @property
    def decode_error(self) -> Type[Exception]:
        from tomlkit.exceptions import ParseError

        return ParseError

    def loads(self, contents: str) -> dict:
        import tomlkit

        return tomlkit.loads(contents)

    def dumps(self, loaded_contents: dict) -> str:
        import tomlkit

        return tomlkit.dumps(loaded_contents)



from pathlib import Path

import pytest

from config_file.utils import Default, split_on_dot


def test_that_split_on_dot_splits_the_line_correctly():
    """
    config_file.utils.split_on_dot

    Ensure splitting a string with dots in it returns an
    array of strings split on those dots.
    """
    assert split_on_dot("section.key.blah") == ["section", "key", "blah"]


def test_that_split_on_dot_can_only_split_on_the_last_one():
    """
    config_file.utils.split_on_dot

    Ensure that the only_last_dot parameter only splits the input
    on the last dot.
    """
    assert split_on_dot("test.input.should.only.split.on.last", only_last_dot=True) == [
        "test.input.should.only.split.on",
        "last",
    ]


def test_that_split_on_dot_raises_error_when_there_is_no_dot():
    """
    config_file.utils.split_on_dot

    Ensure that a ValueError is raised if the input to split_on_dot
    has no dot (.) in it.
    """
    with pytest.raises(ValueError):
        split_on_dot("this line has no dots in it")


def test_that_split_on_dot_can_accept_path_objects():
    """
    config_file.utils.split_on_dot

    Ensure that a pathlib.Path object is able to be
    given to split_on_dot.
    """
    split = split_on_dot(Path("some.file.path.with.dots"))

    assert split == ["some", "file", "path", "with", "dots"]


def test_that_default_is_initialized_correctly():
    """
    config_file.utils.Default

    Ensure that the Default class's value attribute
    is the same value that it was constructed with.
    """
    default = Default(5)
    assert default.value == 5


def test_that_default_has_the_correct_repr_format():
    """
    config_file.utils.Default

    Ensure that when the Default class is coerced to
    a string, it has the correct format and output.
    """
    default = Default("test")
    assert str(default) == "Default Value: test (<class 'str'>)"



from abc import abstractmethod
from typing import Any, Type

from config_file.exceptions import ParsingError
from config_file.nested_lookup import get_occurrence_of_key
from config_file.parsers.abstract_parser import AbstractParser
from config_file.utils import split_on_dot


class BaseParser(AbstractParser):
    def __init__(self, file_contents: str):
        """
        The BaseParser implements the AbstractParser for us, as long as the
        subclasses implement `loads`, `dumps`, and `decode_error` methods.

        Internally, every parser uses the BaseParser so it does not
        have to re-implement common dictionary manipulation logic. However,
        there is nothing about the BaseParser that is a requirement for
        using ConfigFile.

        Args:
            file_contents: The contents of the file to parse.

        Raises:
            ParsingError: If the decode_error is raised.
        """
        super().__init__(file_contents)
        self.__parsed_content = self.parse_file_contents()

    @abstractmethod
    def loads(self, contents: str) -> dict:
        """
        Transforms the file contents into a dictionary.

        Args:
            contents: The file contents to transform.

        Raises:
            self.decode_error: If there is a decoding error
                while attempting to parse the string.

        Returns:
            The contents of a particular file as a dictionary
            we can manipulate.
        """
        raise NotImplementedError

    @abstractmethod
    def dumps(self, loaded_contents: dict) -> str:
        """
        Transform the parsed contents back into a string.

        Args:
            loaded_contents: The file contents from a self.loads call.
                It doesn't necessary need to be from a loads call, could just
                be a valid dictionary for the given format, but that is the
                typical usecase.

        Raises:
            self.decode_error: If there is a decoding error
                while attempting to transform the dict back into a string.

        Returns:
            The dictionary parsed back into a string.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def decode_error(self) -> Type[Exception]:
        """
        The decoding error raised by the subclass on loads/dumps.

        Returns:
            The decoding error
        """
        raise NotImplementedError

    @property
    def parsed_content(self) -> dict:
        return self.__parsed_content

    def __str__(self) -> str:
        return self.dumps(self.__parsed_content)

    def parse_file_contents(self) -> dict:
        """
        Parse the file contents by running the `loads`  method on the module.

        Args:
            module: The module to use to parse the file.
            decode_error: The error that is raised from the module if the file
                cannot be decoded.

        Raises:
            ParsingError: If the decode_error is raised while running `loads`.

        Returns:
            The parsed file as a dictionary.
        """
        try:
            return self.loads(self.file_contents)
        except self.decode_error as error:
            raise ParsingError(error)

    def reset_internal_contents(self, file_contents: str) -> None:
        """
        Reset the file contents and parsed contents of the parser.

        Args:
            file_contents: The new file contents.

        Raises:
            ParsingError: If we're unable to parse the new file contents.
        """
        self.file_contents = file_contents
        self.__parsed_content = self.parse_file_contents()

    def get(self, search_key: str) -> Any:
        """
        Retrieve a key from the parsed content.

        Args:
            search_key: The key to search from in the
                parsed content in a "dot" syntax.
                i.e. "section.key"

        Raises:
            KeyError: If the key or one of the sections
                we're subscripting into does not exist.

        Returns:
            The value of the key we're searching for.
        """
        error_key = None

        try:
            if "." not in search_key:
                error_key = search_key
                return self.__parsed_content[search_key]
            else:
                split_keys = split_on_dot(search_key)

            parsed_content = self.__parsed_content
            for key in split_keys:
                error_key = key
                parsed_content = parsed_content[key]
        except (KeyError, TypeError):
            raise KeyError(
                f"cannot `get` {search_key} because "
                f"{error_key} is not subscriptable."
            )

        return parsed_content

    def set(self, key: str, value: Any) -> None:
        if "." not in key:
            self.__parsed_content[key] = value
            return

        keys = split_on_dot(key)
        parsed_content = self.__parsed_content
        for index, key in enumerate(keys):
            if index == len(keys) - 1:
                parsed_content[key] = value
                return

            try:
                parsed_content = parsed_content[key]
            except KeyError:
                parsed_content[key] = {}
                parsed_content = parsed_content[key]

    def delete(self, key: str) -> None:
        try:
            if "." not in key:
                del self.__parsed_content[key]
            else:
                indexes = split_on_dot(key)
                indexes_length = len(indexes)

                parsed_content = self.__parsed_content
                for index, key in enumerate(indexes):
                    if index == indexes_length - 1:
                        del parsed_content[key]
                        break

                    parsed_content = parsed_content[key]
        except (KeyError, TypeError):
            raise KeyError(f"The specified key '{key}' to delete was not found.")

    def has(self, search_key: str, wild: bool = False) -> bool:
        if wild:
            return get_occurrence_of_key(self.__parsed_content, key=search_key) > 0

        try:
            self.get(search_key)
            return True
        except (KeyError, ParsingError):
            return False



from unittest.mock import patch

import pytest

from config_file.parsers.yaml_parser import YamlParser


def test_that_import_error_is_raised_if_no_ruamelyaml():
    """
    config_file.yaml_parser.YamlParser

    Ensure an import error is raised if ruamel.yaml is not
    installed.
    """
    with pytest.raises(ImportError) as error:
        with patch("builtins.__import__") as mock_import:
            mock_import.side_effect = ImportError
            YamlParser("")

    assert (
        "Install the `yaml` extra first with `pip install config-file[yaml]`."
        in str(error)
    )



import json
from pathlib import Path
from typing import Callable, Tuple

import pytest

from config_file.config_file import ConfigFile
from config_file.config_file_path import ConfigFilePath
from config_file.exceptions import ParsingError
from config_file.parsers.abstract_parser import AbstractParser
from config_file.parsers.ini_parser import IniParser

SUPPORTED_FILE_TYPES = ["ini", "json", "yaml", "toml"]


@pytest.fixture(params=SUPPORTED_FILE_TYPES)
def template_and_config_file(
    request, template_file
) -> Callable[[str], Tuple[Path, ConfigFile]]:
    """
    Fixture that is a wrapper around the template_file
    fixture. Using the function returned, we can retrieve
    a temporary file that has the contents of the specified
    template file as well as a ConfigFile with the template
    already passed in.

    This fixture will make the test function run once for every
    supported file type, each with the corresponding template
    in the different file type.
    """

    def func(template_name: str = "default"):
        template = template_file(request.param, template_name)
        return template, ConfigFile(template)

    return func


@pytest.fixture(params=SUPPORTED_FILE_TYPES)
def templated_config_file(template_and_config_file) -> Callable[[str], ConfigFile]:
    """
    Wrapper around template_and_config_file for when we
    only need the ConfigFile and not the template.
    """

    def func(template_name: str = "default"):
        return template_and_config_file(template_name)[1]

    return func


def test_config_file_init_has_correct_path_and_parser(template_and_config_file):
    """
    config_file.config_file.ConfigFile.__init__

    Ensure that an initialized ConfigFile has the
    correct file path and has a parser that is a AbstractParser.
    """
    template_file, config = template_and_config_file()

    assert config.path == template_file
    assert isinstance(config._ConfigFile__parser, AbstractParser)


@pytest.mark.parametrize("file_type", SUPPORTED_FILE_TYPES)
def test_config_file_raises_error_on_malformed_file(template_file, file_type):
    """
    config_file.config_file.ConfigFile.__init__

    Ensure that a ParsingError is raised on a malformed file.
    """
    with pytest.raises(ParsingError):
        ConfigFile(template_file(file_type, template_name="invalid"))


def test_config_file_expands_home_tildes_in_path():
    """
    config_file.config_file.ConfigFile.__init__

    Ensure that a home tilde (~) in a file path passed
    to ConfigFile is expanded to the full path.
    """
    config_name = "temp.ini"
    config_path = "~/" + config_name

    try:
        Path(config_path).expanduser().touch()
        assert ConfigFile(config_path).path == Path.home() / config_name
    finally:
        Path(config_path).expanduser().unlink()


@pytest.mark.parametrize("path", ["invalid", "invalid.conf"])
def test_config_file_raises_error_on_non_existent_path(path):
    """
    config_file.config_file.ConfigFile.__init__

    Ensure that when a path to a file that does not exist
    that is passed into the constructor, it raises a FileNotFoundError.
    """
    with pytest.raises(FileNotFoundError):
        ConfigFile(path)


def test_config_file_raises_error_if_path_is_directory():
    """
    config_file.config_file.ConfigFile.__init__

    Ensure when a path to a directory is passed into
    the constructor, it raises a ValueError.
    """
    with pytest.raises(ValueError):
        ConfigFile(".")


@pytest.mark.parametrize(
    "key, value",
    [
        ("header_one.list_key", "different value"),
        ("header_one", {}),
        ("header_one.number_key", 1),
    ],
)
def test_set_can_add_to_or_alter_the_file(templated_config_file, key, value):
    """
    config_file.config_file.ConfigFile.set

    Ensure that for the keys we can, we retrieve that same value back
    (with the exception of ini, which should always be a string).
    """
    config = templated_config_file()
    config.set(key, value)

    # configparser treats everything as a string. Make
    # sure we're entering in files in the dict as a str.
    if ConfigFilePath(config.path).extension == "ini":
        if isinstance(value, (int, float)):
            value = str(value)

        assert config.get(key) == value
    else:
        assert config.get(key) == value


def test_config_file_can_save(template_and_config_file):
    """
    config_file.config_file.ConfigFile.save

    Ensure that altering the configuration file and saving it
    writes those changes out to the filesystem.
    """
    template_file, config = template_and_config_file()

    config.set("header_one.number_key", 25)
    config.save()

    assert str(config) == template_file.read_text()


def test_restore_original_can_restore_with_calculated_in_path(
    templated_config_file, template_original_file
):
    """
    config_file.config_file.ConfigFile.restore_original

    Ensure that we can restore the "original" configuration
    file, using the *calculated* path, after altering the
    internal config contents.

    Those internal contents and file should be reset to the
    original file.
    """
    config = templated_config_file()
    original_file = template_original_file(config.path)

    config.set("header_one.number_key", 5)
    config.restore_original()

    assert str(config) == original_file.read_text()
    assert config.path.read_text() == original_file.read_text()


@pytest.mark.parametrize("file_type", SUPPORTED_FILE_TYPES)
def test_restore_original_can_restore_with_passed_in_path(
    template_file, tmp_file, file_type
):
    """
    config_file.config_file.ConfigFile.restore_original

    Ensure that we can restore the "original" configuration
    file, using the *passed* in path, after altering the
    internal config contents.
    """
    config_file = template_file(file_type)
    config = ConfigFile(config_file)

    original_file = tmp_file(f"file.{file_type}", config_file.read_text())
    original_config = ConfigFile(original_file)
    original_config.set("some_new_section.some_new_key", 5)
    original_config.save()

    config.restore_original(original_path=original_file)

    assert str(config) == str(original_config)
    assert config.path.read_text() == original_file.read_text()


def test_restore_original_raises_error_on_missing_file(templated_config_file):
    """
    config_file.config_file.ConfigFile.restore_original

    Ensure that if we try to restore_original without using
    the optional original_path argument but we do not
    have a file at the calculated default path, a FileNotFoundError
    is raised.
    """
    config = templated_config_file()

    with pytest.raises(FileNotFoundError) as error:
        config.restore_original()

    assert "The specified config file" in str(error)
    assert "does not exist" in str(error)


@pytest.mark.parametrize("key", ["header_one", "header_one.number_key"])
def test_has_can_find_sections_and_keys(templated_config_file, key):
    """
    config_file.config_file.ConfigFile.has

    Ensure that when a section or key that exists
    is passed to has(), it can find it and returns True.
    """
    config = templated_config_file()
    assert config.has(key)
    assert key in config


def test_has_returns_false_for_sections_and_keys_that_do_not_exist(
    templated_config_file,
):
    """
    config_file.config_file.ConfigFile.has

    Ensure that when a section or key that does not exist
    is passed to has(), it can't find it and returns False.
    """
    config = templated_config_file()
    assert not config.has("header_one.blah")


def test_has_can_search_for_a_key_anywhere(templated_config_file):
    """
    config_file.config_file.ConfigFile.has

    Ensure that the wild=True keyword argument to has()
    retrieves if we have the key anywhere in the file.
    """
    config = templated_config_file()
    assert config.has("list_key", wild=True)


@pytest.mark.parametrize(
    "key, value",
    [
        ("header_one.number_key", 0),
        ("header_two.list_key", [1, 2, 3]),
        ("header_one", {"number_key": 0}),
    ],
)
def test_get_can_retrieve_existent_keys(templated_config_file, key, value):
    """
    config_file.config_file.ConfigFile.get

    Ensure that we can retrieve a section or key
    and that it is in its native type (the type that
    it was set to).

    The one caveat here is the IniParser. The IniParser
    uses configparser under the hood which simply always
    returns back a string. However, that is not the case
    for JsonParser, YamlParser, or TomlParser.
    """
    config = templated_config_file()

    if isinstance(config._ConfigFile__parser, IniParser):
        if type(value) is not dict:
            # not a section being retrieved
            value = str(value)
        else:
            # section is being retrieved so it should be a dict
            # but the values inside should be strings.
            value = json.loads(json.dumps(value), parse_int=str)

    assert config.get(key) == value


def test_get_can_parse_all_types(templated_config_file):
    """
    config_file.config_file.ConfigFile.get

    Ensure we're able to parse out all the different types that
    may be used in a configuration file with the parse_types option.
    """
    config = templated_config_file(template_name="all_strings")
    output = {"list": [], "dict": {}, "num": 5, "bool": False, "float": 5.5}

    assert config.get("header", parse_types=True) == output


def test_get_can_use_a_default(templated_config_file):
    config = templated_config_file()
    assert config.get("does_not_exist", default="default!") == "default!"


def test_get_raises_a_key_error_on_invalid_key(templated_config_file):
    config = templated_config_file()

    with pytest.raises(KeyError):
        config.get("does_not_exist")


@pytest.mark.parametrize("key", ["header_one.number_key", "header_one"])
def test_delete_can_remove_sections_and_keys(templated_config_file, key):
    """
    config_file.config_file.ConfigFile.delete

    Ensure that we can delete keys and that they are removed
    from the internal contents of the ConfigFile.
    """
    config = templated_config_file()

    config.delete(key)

    assert config.has(key) is False


@pytest.mark.parametrize("key", ["", 0, False, {}, "header_one.does_not_exist"])
def test_delete_raises_key_error_on_invalid_input_or_missing_key(
    templated_config_file, key
):
    """
    config_file.config_file.ConfigFile.delete

    Ensure that a KeyError is raised if we
    tried to provide a value that is not valid
    (like the wrong type so it is an invalid key)
    or one that does not exist.
    """
    config = templated_config_file()

    with pytest.raises(KeyError):
        config.delete(key)


def test_keys_can_be_set_with_array_notation(template_and_config_file):
    """
    config_file.config_file.ConfigFile.__setitem__

    Ensure that the ConfigFile exposes its parsed contents
    using an array notation so we can set keys that way.
    """
    template_file, config = template_and_config_file()

    config["header_one"] = {}
    config.save()

    assert str(config) == template_file.read_text()


def test_keys_can_be_retrieved_with_array_notation(templated_config_file):
    """
    config_file.config_file.ConfigFile.__getitem__

    Ensure that the ConfigFile exposes its parsed contents
    using an array notation so we can retrieve keys that way.
    """
    config = templated_config_file()
    assert int(config["header_one"]["number_key"]) == 0


def test_keys_can_be_deleted_with_array_notation(templated_config_file):
    """
    config_file.config_file.ConfigFile.__delitem__

    Ensure that the ConfigFile exposes its parsed contents
    using an array notation so we can delete keys that way.
    """
    config = templated_config_file()
    del config["header_one"]

    with pytest.raises(KeyError):
        config["header_one"]


def test_the_config_file_can_be_stringified(template_and_config_file):
    """
    config_file.config_file.ConfigFile.__str__
    config_file.config_file.ConfigFile.__repr__

    Ensure that the stringified version of the config file
    is the string version of the file and that the repr also
    includes the file path.
    """
    template, config = template_and_config_file()
    assert template.read_text() == str(config)
    assert f"{str(template)}\n\n{str(config)}" == repr(config)



"""This module allows you to parse a string into its native type."""
import ast
import re
from typing import Any


def parse_value(value: Any) -> Any:
    """Parse a value into native type.

    parse_value('5') -> 5   (int)
    parse_value('-5') -> -5 (int)

    parse_value('5.52') -> 5.52   (float)
    parse_value('-53.2') -> -53.2 (float)

    parse_value('blahfafda') -> 'blahfafda' (string)

    parse_value('true') -> True (bool)
    parse_value('FALSE') -> False (bool)

    parse_value("  [ [ [ [] ]] ]  ") -> [[[[]]]] (list)

    Args:
        value: The value to parse.

    Returns:
        The parsed value, if it could be parsed
        as a dictionary, list, int, float, or bool.
        Otherwise, it simply returns the original
        passed in value.
    """
    if isinstance(value, dict):
        parsed = {}

        for item in value:
            parsed[item] = parse_value(value[item])

        return parsed

    if isinstance(value, list):
        parsed = []

        for item in value:
            parsed.append(parse_value(item))

        return parsed

    if can_be_parsed_as_int(value):
        return int(value.strip()) if isinstance(value, str) else int(value)

    if can_be_parsed_as_float(value):
        return float(value.strip()) if isinstance(value, str) else float(value)

    if can_be_parsed_as_bool(value):
        return bool(value) if isinstance(value, bool) else bool(strtobool(value))

    if can_be_parsed_as_dict(value):
        return parse_value(ast.literal_eval(value.strip()))

    if can_be_parsed_as_list(value):
        return parse_value(ast.literal_eval(value.strip()))

    return value


def can_be_parsed_as_int(value: Any) -> bool:
    """Check whether a value can be parsed as a integer.

    can_be_parsed_as_int(5) -> True
    can_be_parsed_as_int("5") -> True
    can_be_parsed_as_int("-5") -> True
    can_be_parsed_as_int(True) -> False

    Args:
        value: The value to check if it can be parsed as an integer.

    Returns:
        True if we can parse the value as a integer. False otherwise.
    """
    if isinstance(value, int) and not isinstance(value, bool):
        return True

    if not isinstance(value, str):
        return False

    value = value.strip()

    if value.startswith("-"):
        return value[1:].isdigit()

    return value.isdigit()


def can_be_parsed_as_float(value: Any) -> bool:
    """Checks whether a value can be parsed as a float.

    can_be_parsed_as_float(45) -> False
    can_be_parsed_as_float(1.1) -> True
    can_be_parsed_as_float("1.1") -> True
    can_be_parsed_as_float("-1.1") -> True
    can_be_parsed_as_float(".1") -> True

    Args:
        value: The value to be parsed.

    Returns:
        True if the value can be parsed as a float. False otherwise.
    """
    FLOAT_REGEX = r"^\d*\.\d+$"

    if isinstance(value, float):
        return True

    if not isinstance(value, str):
        return False

    value = value.strip()

    if value.startswith("-"):
        value = value[1:]

    return bool(re.match(FLOAT_REGEX, value))


def can_be_parsed_as_bool(value: Any) -> bool:
    """Checks whether a value can be parsed as a boolean.

    can_be_parsed_as_bool(True) -> True
    can_be_parsed_as_bool("true") -> True
    can_be_parsed_as_bool("false") -> True
    can_be_parsed_as_bool("TRUE") -> True
    can_be_parsed_as_bool(0) -> False

    Args:
        value: The value to be parsed.

    Returns:
        True if the value can be parsed as a boolean. False otherwise.
    """
    if isinstance(value, bool):
        return True

    if not isinstance(value, str):
        return False

    value = value.lower().strip()

    return value == "true" or value == "false"


# distutils strtobool backport: https://github.com/yukinarit/strtobool
def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).
    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return 1
    elif val in ("n", "no", "f", "false", "off", "0"):
        return 0
    else:
        raise ValueError("invalid truth value %r" % (val,))


def can_be_parsed_as_dict(value: Any) -> bool:
    """Checks whether a value can be parsed as a dictionary.

    can_be_parsed_as_dict({}) -> True
    can_be_parsed_as_dict("{}") -> True
    can_be_parsed_as_dict("  {  }  ") -> True
    can_be_parsed_as_dict("{ 'hello': 5 }") -> True
    can_be_parsed_as_dict([]]) -> False
    can_be_parsed_as_dict(" { invalid } ") -> False

    Args:
        value: The value to be parsed.

    Returns:
        True if the value can be parsed as a dictionary. False otherwise.
    """
    if isinstance(value, dict):
        return True

    if not isinstance(value, str):
        return False

    value = value.strip()

    if value[0] == "{" and value[-1] == "}":
        try:
            ast.literal_eval(value)
            return True
        except Exception:
            return False

    return False


def can_be_parsed_as_list(value: Any) -> bool:
    """Checks whether a value can be parsed as a list.

    can_be_parsed_as_list([]) -> True
    can_be_parsed_as_list([1, 2]) -> True
    can_be_parsed_as_list("[]") -> True
    can_be_parsed_as_list("  []  ") -> True
    can_be_parsed_as_list("[1, 2]") -> True
    can_be_parsed_as_list({}) -> False
    can_be_parsed_as_list(" [ invalid ] ") -> False

    Args:
        value: The value to be parsed.

    Returns:
        True if the value can be parsed as a list. False otherwise.
    """
    if isinstance(value, list):
        return True

    if not isinstance(value, str):
        return False

    value = value.strip()

    if value[0] == "[" and value[-1] == "]":
        try:
            ast.literal_eval(value)
            return True
        except Exception:
            return False

    return False



import pytest

from config_file.parsers.abstract_parser import AbstractParser
from config_file.parsers.base_parser import BaseParser


class ConcreteAbstractParser(AbstractParser):
    """
    ConcreteParser is a concrete implementation of the
    AbstractParser that just calls the AbstractParser's
    methods so we can ensure they raise a NotImplementedError.
    """

    def __init__(self, file_contents):
        super().__init__(file_contents)

    def get(self, key):
        super().get(key)

    def set(self, key, value):
        super().set(key, value)

    def delete(self, section_key):
        super().delete(section_key)

    def has(self, section_key: str, wild: bool = False):
        super().has(section_key)

    def __str__(self):
        super().__str__()

    def reset_internal_contents(self, file_contents: str) -> None:
        super().reset_internal_contents(file_contents)

    def parsed_content(self) -> dict:
        super().parsed_content()


def test_that_abstract_parser_can_not_be_instantiated():
    """
    config_file.abstract_parser.AbstractParser

    Ensure the AbstractParser raises a TypeError when it is
    instantiated.
    """
    with pytest.raises(TypeError):
        AbstractParser("")


def test_that_base_parser_can_not_be_instantiated():
    """
    config_file.base_parser.BaseParser

    Ensure the BaseParser raises a TypeError when it is
    instantiated.
    """
    with pytest.raises(TypeError):
        BaseParser("")


@pytest.mark.parametrize(
    "function, args",
    [
        (ConcreteAbstractParser("").get, ("",)),
        (ConcreteAbstractParser("").set, ("", "")),
        (ConcreteAbstractParser("").delete, ("",)),
        (ConcreteAbstractParser("").__str__, None),
        (ConcreteAbstractParser("").has, ("",)),
        (ConcreteAbstractParser("").reset_internal_contents, ("",)),
        (ConcreteAbstractParser("").parsed_content, None),
    ],
)
def test_that_abstract_and_base_parser_raises_not_implemented_errors(function, args):
    """
    config_file.abstract_parser.AbstractParser
    config_file.base_parser.BaseParser

    Ensure the AbstractParser and BaseParser raises a NotImplementedError
    for every method in it.
    """
    with pytest.raises(NotImplementedError):
        if args is None:
            function()
        else:
            function(*args)



from io import StringIO
from typing import Type

from .base_parser import BaseParser


class YamlParser(BaseParser):
    def __init__(self, file_contents: str):
        try:
            super().__init__(file_contents)
        except ImportError:
            raise ImportError(
                "It doesn't appear `ruamel.yaml` is installed, but a yaml "
                "file was attempted to be used. Install the `yaml` "
                "extra first with `pip install config-file[yaml]`."
            )

    @property
    def decode_error(self) -> Type[Exception]:
        from ruamel.yaml import YAMLError

        return YAMLError

    def loads(self, contents: str) -> dict:
        from ruamel.yaml import YAML

        return YAML().load(contents)

    def dumps(self, loaded_contents: dict) -> str:
        from ruamel.yaml import YAML

        buffer = StringIO()
        YAML().dump(loaded_contents, buffer)
        return buffer.getvalue()



class ParsingError(ValueError):
    """
    Unable to parse the configuration file.

    Since every module that parses the different
    files has it's own decode error, this standardizes
    it for the package to only raise a ParsingError
    if a error occurs in those instances.
    """



from unittest import TestCase

from config_file.nested_lookup import (
    get_all_keys,
    get_occurrence_of_key,
    get_occurrence_of_value,
    get_occurrences_and_values,
    nested_alter,
    nested_delete,
    nested_lookup,
    nested_update,
)


class TestNestedLookup(TestCase):
    def setUp(self):
        self.subject_dict = {"a": 1, "b": {"d": 100}, "c": {"d": 200}}
        self.subject_dict2 = {
            "name": "Russell Ballestrini",
            "email_address": "test1@example.com",
            "other": {
                "secondary_email": "test2@example.com",
                "EMAIL_RECOVERY": "test3@example.com",
                "email_address": "test4@example.com",
            },
        }
        self.subject_dict3 = {
            "build_version": {
                "model_name": "MacBook Pro",
                "build_version": {
                    "processor_name": "Intel Core i7",
                    "processor_speed": "2.7 GHz",
                    "core_details": {
                        "build_version": "4",
                        "l2_cache(per_core)": "256 KB",
                    },
                },
                "number_of_cores": "4",
                "memory": "256 KB",
            },
            "os_details": {"product_version": "10.13.6", "build_version": "17G65"},
            "name": "Test",
            "date": "YYYY-MM-DD HH:MM:SS",
        }
        self.subject_dict4 = {
            1: "a",
            2: {"b": 44, "C": 55},
            3: "d",
            4: "e",
            "6776": "works",
        }

    def test_nested_lookup(self):
        results = nested_lookup("d", self.subject_dict)
        self.assertEqual(2, len(results))
        self.assertIn(100, results)
        self.assertIn(200, results)
        self.assertSetEqual({100, 200}, set(results))

    def test_nested_lookup_wrapped_in_list(self):
        results = nested_lookup("d", [{}, self.subject_dict, {}])
        self.assertEqual(2, len(results))
        self.assertIn(100, results)
        self.assertIn(200, results)
        self.assertSetEqual({100, 200}, set(results))

    def test_nested_lookup_wrapped_in_list_in_dict_in_list(self):
        results = nested_lookup("d", [{}, {"H": [self.subject_dict]}])
        self.assertEqual(2, len(results))
        self.assertIn(100, results)
        self.assertIn(200, results)
        self.assertSetEqual({100, 200}, set(results))

    def test_nested_lookup_wrapped_in_list_in_list(self):
        results = nested_lookup("d", [{}, [self.subject_dict, {}]])
        self.assertEqual(2, len(results))
        self.assertIn(100, results)
        self.assertIn(200, results)
        self.assertSetEqual({100, 200}, set(results))

    def test_nested_lookup_key_is_non_str(self):
        results = nested_lookup(key=4, document=self.subject_dict4)
        self.assertIn("e", results)

    def test_wild_nested_lookup(self):
        results = nested_lookup(key="mail", document=self.subject_dict2, wild=True)
        self.assertEqual(4, len(results))
        self.assertIn("test1@example.com", results)
        self.assertIn("test2@example.com", results)
        self.assertIn("test3@example.com", results)

    def test_wild_nested_lookup_integer_keys_in_document(self):
        results = nested_lookup(key="c", document=self.subject_dict4, wild=True)
        self.assertIn(55, results)

    def test_wild_nested_lookup_integer_key_as_substring(self):
        # test that wild works converts integers into strings before substring matching.
        results = nested_lookup(key=77, document=self.subject_dict4, wild=True)
        self.assertIn("works", results)

    def test_wild_with_keys_nested_lookup(self):
        matches = nested_lookup(
            key="mail", document=self.subject_dict2, wild=True, with_keys=True
        )
        self.assertEqual(3, len(matches))
        self.assertIn("email_address", matches)
        self.assertIn("secondary_email", matches)
        self.assertIn("EMAIL_RECOVERY", matches)
        self.assertSetEqual(
            {"test1@example.com", "test4@example.com"}, set(matches["email_address"])
        )
        self.assertIn("test2@example.com", matches["secondary_email"])

    def test_nested_lookup_with_keys(self):
        matches = nested_lookup("d", self.subject_dict, with_keys=True)
        self.assertIn("d", matches)
        self.assertEqual(2, len(matches["d"]))
        self.assertSetEqual({100, 200}, set(matches["d"]))

    def test_after_key_is_found(self):
        result = nested_lookup(key="build_version", document=self.subject_dict3)
        self.assertEqual(4, len(result))
        self.assertIn("4", result)
        self.assertIn("17G65", result)
        match1 = {
            "processor_name": "Intel Core i7",
            "processor_speed": "2.7 GHz",
            "core_details": {"build_version": "4", "l2_cache(per_core)": "256 KB"},
        }
        self.assertIn(match1, result)
        match2 = {
            "build_version": {
                "processor_name": "Intel Core i7",
                "processor_speed": "2.7 GHz",
                "core_details": {"build_version": "4", "l2_cache(per_core)": "256 KB"},
            },
            "memory": "256 KB",
            "model_name": "MacBook Pro",
            "number_of_cores": "4",
        }
        self.assertIn(match2, result)


class TestGetAllKeys(TestCase):
    def setUp(self):
        self.sample1 = {
            "hardware_details": {
                "model_name": "MacBook Pro",
                "processor_details": {
                    "processor_name": "Intel Core i7",
                    "processor_speed": "2.7 GHz",
                    "core_details": {
                        "total_numberof_cores": "4",
                        "l2_cache(per_core)": "256 KB",
                    },
                },
                "total_number_of_cores": "4",
                "memory": "16 GB",
            },
            "os_details": {"product_version": "10.13.6", "build_version": "17G65"},
            "name": "Test",
            "date": "YYYY-MM-DD HH:MM:SS",
        }
        self.sample2 = {
            "hardware_details": {
                "model_name": "MacBook Pro",
                "processor_details": [
                    {
                        "processor_name": "Intel Core i7",
                        "processor_speed": "2.7 GHz",
                        "core_details": {
                            "total_numberof_cores": "4",
                            "l2_cache(per_core)": "256 KB",
                        },
                    }
                ],
                "total_number_of_cores": "4",
                "memory": "16 GB",
            }
        }
        self.sample3 = {
            "hardware_details": {
                "model_name": "MacBook Pro",
                "processor_details": [
                    {"processor_name": "Intel Core i7", "processor_speed": "2.7 GHz"},
                    {"total_numberof_cores": "4", "l2_cache(per_core)": "256 KB"},
                ],
                "total_number_of_cores": "4",
                "memory": "16 GB",
            }
        }
        self.sample4 = {
            "values": [
                {
                    "checks": [
                        {
                            "monitoring_zones": [
                                "mzdfw",
                                "mzfra",
                                "mzhkg",
                                "mziad",
                                "mzlon",
                                "mzord",
                                "mzsyd",
                            ]
                        }
                    ]
                }
            ]
        }
        self.sample5 = [
            {
                "listings": [
                    {
                        "name": "title",
                        "postcode": "postcode",
                        "full_address": "fulladdress",
                        "city": "city",
                        "lat": "latitude",
                        "lng": "longitude",
                    }
                ]
            }
        ]

    def test_sample_data1(self):
        result = get_all_keys(self.sample1)
        self.assertEqual(15, len(result))
        keys_to_verify = [
            "model_name",
            "core_details",
            "l2_cache(per_core)",
            "build_version",
            "date",
        ]
        for key in keys_to_verify:
            self.assertIn(key, result)

    def test_sample_data2(self):
        result = get_all_keys(self.sample2)
        self.assertEqual(10, len(result))
        keys_to_verify = [
            "hardware_details",
            "processor_speed",
            "total_numberof_cores",
            "memory",
        ]
        for key in keys_to_verify:
            self.assertIn(key, result)

    def test_sample_data3(self):
        result = get_all_keys(self.sample3)
        self.assertEqual(9, len(result))
        keys_to_verify = [
            "processor_details",
            "processor_name",
            "l2_cache(per_core)",
            "total_number_of_cores",
        ]
        for key in keys_to_verify:
            self.assertIn(key, result)

    def test_sample_data4(self):
        result = get_all_keys(self.sample4)
        self.assertEqual(3, len(result))
        keys_to_verify = ["values", "checks", "monitoring_zones"]
        for key in keys_to_verify:
            self.assertIn(key, result)

    def test_sample_data5(self):
        result = get_all_keys(self.sample5)
        self.assertEqual(7, len(result))
        keys_to_verify = [
            "listings",
            "name",
            "postcode",
            "full_address",
            "city",
            "lat",
            "lng",
        ]
        for key in keys_to_verify:
            self.assertIn(key, result)


class TestGetOccurrence(TestCase):
    def setUp(self):
        self.sample1 = {
            "build_version": {
                "model_name": "MacBook Pro",
                "build_version": {
                    "processor_name": "Intel Core i7",
                    "processor_speed": "2.7 GHz",
                    "core_details": {
                        "build_version": "4",
                        "l2_cache(per_core)": "256 KB",
                    },
                },
                "number_of_cores": "4",
                "memory": "256 KB",
            },
            "os_details": {"product_version": "10.13.6", "build_version": "17G65"},
            "name": "Test",
            "date": "YYYY-MM-DD HH:MM:SS",
        }
        self.sample2 = {
            "hardware_details": {
                "model_name": "MacBook Pro",
                "processor_details": [
                    {
                        "processor_name": "4",
                        "processor_speed": "2.7 GHz",
                        "core_details": {
                            "total_numberof_cores": "4",
                            "l2_cache(per_core)": "256 KB",
                        },
                    }
                ],
                "total_number_of_cores": "4",
                "memory": "16 GB",
            }
        }
        self.sample3 = {
            "hardware_details": {
                "model_name": "MacBook Pro",
                "processor_details": [
                    {"total_number_of_cores": "4", "processor_speed": "2.7 GHz"},
                    {"total_number_of_cores": "4", "l2_cache(per_core)": "256 KB"},
                ],
                "total_number_of_cores": "4",
                "memory": "16 GB",
            }
        }
        self.sample4 = {
            "values": [
                {
                    "checks": [
                        {
                            "monitoring_zones": [
                                "mzdfw",
                                "mzfra",
                                "mzhkg",
                                "mziad",
                                "mzlon",
                                "mzord",
                                "mzsyd",
                            ]
                        }
                    ]
                }
            ]
        }

        self.sample5 = {
            "hardware_details": {
                "model_name": "MacBook Pro",
                "total_number_of_cores": 0,
                "memory": False,
            }
        }

        self.sample6 = [
            {
                "processor_name": "4",
                "processor_speed": "2.7 GHz",
                "core_details": {
                    "total_numberof_cores": "4",
                    "l2_cache(per_core)": "256 KB",
                },
            }
        ]

        self.sample7 = [
            {
                "processor_name": "4",
                "processor_speed": "2.7 GHz",
                "core_details": {
                    "total_numberof_cores": "4",
                    "l2_cache(per_core)": "256 KB",
                },
            },
            {
                "processor_name": "4",
                "processor_speed": "2.7 GHz",
                "core_details": {
                    "total_numberof_cores": "4",
                    "l2_cache(per_core)": "256 KB",
                },
            },
        ]

    def test_sample_data1(self):
        result = get_occurrence_of_key(self.sample1, "build_version")
        self.assertEqual(4, result)
        result = get_occurrence_of_value(self.sample1, "256 KB")
        self.assertEqual(2, result)

    def test_sample_data2(self):
        result = get_occurrence_of_key(self.sample2, "core_details")
        self.assertEqual(1, result)
        result = get_occurrence_of_value(self.sample2, "4")
        self.assertEqual(3, result)

    def test_sample_data3(self):
        result = get_occurrence_of_key(self.sample3, "total_number_of_cores")
        self.assertEqual(3, result)
        result = get_occurrence_of_value(self.sample3, "4")
        self.assertEqual(3, result)

    def test_sample_data4(self):
        result = get_occurrence_of_key(self.sample4, "checks")
        self.assertEqual(1, result)
        result = get_occurrence_of_value(self.sample4, "mziad")
        self.assertEqual(1, result)
        # Add one more value in key "monitoring_zones" and verify
        self.sample4["values"][0]["checks"][0]["monitoring_zones"].append("mziad")
        self.assertEqual(2, get_occurrence_of_value(self.sample4, "mziad"))

    def test_sample_data5(self):
        self.assertEqual(
            1, get_occurrence_of_key(self.sample5, "total_number_of_cores")
        )
        self.assertEqual(1, get_occurrence_of_key(self.sample5, "memory"))
        # Add key 'memory' and verify
        self.sample5["memory"] = 0
        self.assertEqual(2, get_occurrence_of_key(self.sample5, "memory"))

    def test_sample_data6(self):
        value = "4"
        result = get_occurrences_and_values(self.sample6, value)
        self.assertEqual(2, result[value]["occurrences"])
        self.assertEqual(2, len(result[value]["values"]))

    def test_sample_data7(self):
        value = "2.7 GHz"
        result = get_occurrences_and_values(self.sample6, value)
        self.assertEqual(1, result[value]["occurrences"])
        self.assertEqual(1, len(result[value]["values"]))

    def test_sample_data8(self):
        value = "4"
        result = get_occurrences_and_values(self.sample7, value)
        self.assertEqual(4, result[value]["occurrences"])
        self.assertEqual(4, len(result[value]["values"]))

    def test_sample_data9(self):
        value = "5"
        result = get_occurrences_and_values(self.sample7, value)
        self.assertEqual(0, result[value]["occurrences"])
        self.assertEqual(0, len(result[value]["values"]))


class BaseLookUpApi(TestCase):
    def setUp(self):
        self.sample_data1 = {
            "build_version": {
                "model_name": "MacBook Pro",
                "build_version": {
                    "processor_name": "Intel Core i7",
                    "processor_speed": "2.7 GHz",
                    "core_details": {
                        "build_version": "4",
                        "l2_cache(per_core)": "256 KB",
                    },
                },
                "number_of_cores": "4",
                "memory": "256 KB",
            },
            "os_details": {"product_version": "10.13.6", "build_version": "17G65"},
            "name": "Test",
            "date": "YYYY-MM-DD HH:MM:SS",
        }

        self.sample_data2 = {
            "hardware_details": {
                "model_name": "MacBook Pro",
                "processor_details": [
                    {"processor_name": "Intel Core i7", "processor_speed": "2.7 GHz"},
                    {"total_number_of_cores": "4", "l2_cache(per_core)": "256 KB"},
                ],
                "total_number_of_cores": "5",
                "memory": "16 GB",
            }
        }

        self.sample_data3 = {
            "values": [
                {
                    "checks": [
                        {
                            "monitoring_zones": [
                                "mzdfw",
                                "mzfra",
                                "mzhkg",
                                "mziad",
                                "mzlon",
                                "mzord",
                                "mzsyd",
                            ]
                        }
                    ]
                }
            ]
        }

        self.sample_data4 = {
            "modelversion": "1.1.0",
            "vorgangsID": "1",
            "versorgungsvorschlagDatum": 1510558834978,
            "eingangsdatum": 1510558834978,
            "plz": 82269,
            "vertragsteile": [
                {
                    "typ": "1",
                    "beitragsDaten": {
                        "endalter": 85,
                        "brutto": 58.76,
                        "netto": 58.76,
                        "zahlungsrhythmus": "MONATLICH",
                        "plz": 86899,
                    },
                    "beginn": 1512082800000,
                    "lebenslang": "True",
                    "ueberschussverwendung": {
                        "ueberschussverwendung": "2",
                        "indexoption": "3",
                    },
                    "deckung": [
                        {
                            "typ": "2",
                            "art": "1",
                            "leistung": {"value": 7500242424.0, "einheit": "2"},
                            "leistungsRhythmus": "1",
                        }
                    ],
                    "zuschlagNachlass": [],
                },
                {
                    "typ": "1",
                    "beitragsDaten": {
                        "endalter": 85,
                        "brutto": 0.6,
                        "netto": 0.6,
                        "zahlungsrhythmus": "1",
                    },
                    "zuschlagNachlass": [],
                },
            ],
        }


class TestNestedDelete(BaseLookUpApi):
    def test_sample_data1(self):
        result = {
            "os_details": {"product_version": "10.13.6"},
            "name": "Test",
            "date": "YYYY-MM-DD HH:MM:SS",
        }
        self.assertEqual(result, nested_delete(self.sample_data1, "build_version"))

    def test_sample_data2(self):
        result = {
            "hardware_details": {
                "model_name": "MacBook Pro",
                "total_number_of_cores": "5",
                "memory": "16 GB",
            }
        }
        self.assertEqual(result, nested_delete(self.sample_data2, "processor_details"))

    def test_sample_data3(self):
        result = {"values": [{"checks": [{}]}]}
        self.assertEqual(result, nested_delete(self.sample_data3, "monitoring_zones"))


class TestNestedUpdate(BaseLookUpApi):
    def test_sample_data1(self):
        result = {
            "build_version": "Test1",
            "os_details": {"product_version": "10.13.6", "build_version": "Test1"},
            "name": "Test",
            "date": "YYYY-MM-DD HH:MM:SS",
        }
        self.assertEqual(
            result, nested_update(self.sample_data1, "build_version", "Test1")
        )

    def test_sample_data1_list_input_treat_list_as_element_true(self):
        result = {
            "build_version": ["Test5", "Test6", "Test7"],
            "os_details": {
                "product_version": "10.13.6",
                "build_version": ["Test5", "Test6", "Test7"],
            },
            "name": "Test",
            "date": "YYYY-MM-DD HH:MM:SS",
        }

        self.assertEqual(
            result,
            nested_update(
                self.sample_data1,
                "build_version",
                ["Test5", "Test6", "Test7"],
                treat_as_element=True,
            ),
        )

    def test_nested_update_in_place_false(self):
        """
        ested_update should mutate and return a copy of the original document
        """
        before_id = id(self.sample_data1)
        result = nested_update(
            self.sample_data1, "build_version", "Test2", in_place=False
        )
        after_id = id(result)
        # the object ids should _not_ match.
        self.assertNotEqual(before_id, after_id)

    def test_nested_update_in_place_true(self):
        """
        nested_update should mutate and return the original document
        """
        before_id = id(self.sample_data1)
        result = nested_update(
            self.sample_data1, "build_version", "Test2", in_place=True
        )
        after_id = id(result)
        # the object ids should match.
        self.assertEqual(before_id, after_id)

    def test_nested_update_in_place_true_list_input(self):
        doc = self.sample_data4
        # get all instances of the given element
        findings = nested_lookup("plz", doc, False, True)
        # alter those instances
        updated_findings = list()
        for key, val in findings.items():
            for elem in val:
                updated_findings.append(elem + 300)
        # update those instances with the altered results
        doc_updated = nested_update(
            doc, "plz", updated_findings, treat_as_element=False
        )
        elem1 = doc_updated["plz"]  # 85269
        # 87199
        elem2 = doc_updated["vertragsteile"][0]["beitragsDaten"]["plz"]
        self.assertEqual(elem1, 82569)
        self.assertEqual(elem2, 87199)

    def test_nested_update_in_place_false_list_input(self):
        doc = self.sample_data4
        # get all instances of the given element
        findings = nested_lookup("plz", doc, False, True)
        # alter those instances
        updated_findings = list()
        for key, val in findings.items():
            for elem in val:
                updated_findings.append(elem + 300)
        # update those instances with the altered results
        doc_updated = nested_update(
            doc, "plz", updated_findings, in_place=False, treat_as_element=False
        )
        elem1 = doc_updated["plz"]
        elem2 = doc_updated["vertragsteile"][0]["beitragsDaten"]["plz"]
        self.assertEqual(elem1, 82569)
        self.assertEqual(elem2, 87199)

    def test_nested_update_in_place_false_list_input_as_element_false(self):
        doc = self.sample_data4
        # get all instances of the given element
        list_input = [1, 2, 3, 4, 5]
        # update those instances with the altered results
        doc_updated = nested_update(
            doc, "plz", list_input, in_place=False, treat_as_element=False
        )
        elem1 = doc_updated["plz"]
        elem2 = doc_updated["vertragsteile"][0]["beitragsDaten"]["plz"]
        # should not work without specifying "treat_list_as_element = True"
        # in nested_update
        self.assertNotEqual(elem1, list_input)
        self.assertNotEqual(elem2, list_input)

    def test_nested_update_in_place_false_list_input_as_element_true(self):
        doc = self.sample_data4
        # get all instances of the given element
        list_input = [1, 2, 3, 4, 5]
        # update those instances with the altered results
        doc_updated = nested_update(
            doc, "plz", list_input, in_place=False, treat_as_element=True
        )
        elem1 = doc_updated["plz"]
        elem2 = doc_updated["vertragsteile"][0]["beitragsDaten"]["plz"]
        # should not work without specifying "treat_list_as_element = True"
        # in nested_update
        self.assertEqual(elem1, list_input)
        self.assertEqual(elem2, list_input)

    def test_nested_update_in_place_true_list_input_as_element_false(self):
        doc = self.sample_data4
        # get all instances of the given element
        list_input = [1, 2, 3, 4, 5]
        # update those instances with the altered results
        doc_updated = nested_update(
            doc, "plz", list_input, in_place=True, treat_as_element=False
        )
        elem1 = doc_updated["plz"]
        elem2 = doc_updated["vertragsteile"][0]["beitragsDaten"]["plz"]
        # should not work without specifying "treat_list_as_element = True"
        # in nested_update
        self.assertNotEqual(elem1, list_input)
        self.assertNotEqual(elem2, list_input)

    def test_nested_update_in_place_true_list_input_as_element_true(self):
        doc = self.sample_data4
        # get all instances of the given element
        list_input = [1, 2, 3, 4, 5]
        # update those instances with the altered results
        doc_updated = nested_update(
            doc, "plz", list_input, in_place=True, treat_as_element=True
        )
        elem1 = doc_updated["plz"]
        elem2 = doc_updated["vertragsteile"][0]["beitragsDaten"]["plz"]
        # should not work without specifying "treat_list_as_element = True"
        # in nested_update
        self.assertEqual(elem1, list_input)
        self.assertEqual(elem2, list_input)

    def test_nested_delete_in_place_false(self):
        """
        nested_delete should mutate and return a copy of the original document
        """
        before_id = id(self.sample_data1)
        result = nested_delete(self.sample_data1, "build_version", in_place=False)
        after_id = id(result)
        # the object ids should _not_ match.
        self.assertNotEqual(before_id, after_id)

    def test_nested_delete_in_place_true(self):
        """nested_delete should mutate and return the original document"""
        before_id = id(self.sample_data1)
        result = nested_delete(self.sample_data1, "build_version", in_place=True)
        after_id = id(result)
        # the object ids should match.
        self.assertEqual(before_id, after_id)

    def test_nested_update_taco_for_example(self):
        document = [{"taco": 42}, {"salsa": [{"burrito": {"taco": 69}}]}]

        updated_document = nested_update(
            document, "taco", [100, 200], treat_as_element=False
        )

        self.assertEqual(updated_document[0]["taco"], 100)
        # The multi-update version only works for scalar input,
        # if you need to adress a list of dicts, you have to
        # manually iterate over those and pass them to nested_update
        # one by one
        self.assertEqual(updated_document[1]["salsa"][0]["burrito"]["taco"], 200)

    def test_nested_update_raise_error(self):
        doc = self.sample_data4
        # get all instances of the given element
        list_input = 1
        # update those instances with the altered results
        self.assertRaises(
            Exception,
            nested_update,
            doc,
            "plz",
            list_input,
            in_place=True,
            treat_as_element=False,
        )

    def test_sample_data2(self):
        result = {
            "hardware_details": {
                "model_name": "MacBook Pro",
                "processor_details": {"test_key1": "test_value1"},
                "total_number_of_cores": "5",
                "memory": "16 GB",
            }
        }
        self.assertEqual(
            result,
            nested_update(
                self.sample_data2, "processor_details", {"test_key1": "test_value1"}
            ),
        )

    def test_sample_data3(self):
        result = {"values": [{"checks": {"key1": ["value1"], "key2": "value2"}}]}
        self.assertEqual(
            result,
            nested_update(
                self.sample_data3, "checks", {"key1": ["value1"], "key2": "value2"}
            ),
        )

    def test_sample_data4(self):
        result = {
            "modelversion": {"key1": ["value1"], "key2": "value2"},
            "vorgangsID": "1",
            "versorgungsvorschlagDatum": 1510558834978,
            "eingangsdatum": 1510558834978,
            "plz": 82269,
            "vertragsteile": [
                {
                    "typ": "1",
                    "beitragsDaten": {
                        "endalter": 85,
                        "brutto": 58.76,
                        "netto": 58.76,
                        "zahlungsrhythmus": "MONATLICH",
                        "plz": 86899,
                    },
                    "beginn": 1512082800000,
                    "lebenslang": "True",
                    "ueberschussverwendung": {
                        "ueberschussverwendung": "2",
                        "indexoption": "3",
                    },
                    "deckung": [
                        {
                            "typ": "2",
                            "art": "1",
                            "leistung": {"value": 7500242424.0, "einheit": "2"},
                            "leistungsRhythmus": "1",
                        }
                    ],
                    "zuschlagNachlass": [],
                },
                {
                    "typ": "1",
                    "beitragsDaten": {
                        "endalter": 85,
                        "brutto": 0.6,
                        "netto": 0.6,
                        "zahlungsrhythmus": "1",
                    },
                    "zuschlagNachlass": [],
                },
            ],
        }
        self.assertEqual(
            result,
            nested_update(
                self.sample_data4,
                "modelversion",
                {"key1": ["value1"], "key2": "value2"},
            ),
        )


class TestNestedAlter(BaseLookUpApi):
    def test_nested_alter_in_place_true(self):
        # callback functions
        def callback(data):
            return str(data) + "###"

        doc_updated = nested_alter(
            self.sample_data4, "vorgangsID", callback, in_place=True
        )

        vorgangsid = doc_updated["vorgangsID"]

        self.assertEqual(vorgangsid, "1###")

    def test_nested_alter_in_place_false(self):
        # callback functions
        def callback(data):
            return str(data) + "###"

        doc_updated = nested_alter(
            self.sample_data4, "vorgangsID", callback, in_place=False
        )

        vorgangsid = doc_updated["vorgangsID"]

        # should not work without specifying
        # "treat_list_as_element = True" in nested_update
        self.assertEqual(vorgangsid, "1###")

    def test_nested_alter_list_input_in_place_true(self):
        # callback functions
        def callback(data):
            return str(data) + "###"

        doc_updated = nested_alter(
            self.sample_data4, ["plz", "vorgangsID"], callback, in_place=True
        )

        plz1 = doc_updated["plz"]
        plz2 = doc_updated["vertragsteile"][0]["beitragsDaten"]["plz"]
        vorgangsid = doc_updated["vorgangsID"]

        # should not work without specifying
        # "treat_list_as_element = True" in nested_update
        self.assertEqual(plz1, "82269###")
        self.assertEqual(plz2, "86899###")
        self.assertEqual(vorgangsid, "1###")

    def test_nested_alter_list_input_with_args_in_place_true(self):
        # callback functions
        def callback(data, str1, str2):
            return str(data) + str1 + str2

        doc_updated = nested_alter(
            self.sample_data4,
            ["plz", "vorgangsID"],
            callback,
            function_parameters=["abc", "def"],
            in_place=True,
        )

        plz1 = doc_updated["plz"]
        plz2 = doc_updated["vertragsteile"][0]["beitragsDaten"]["plz"]
        vorgangsid = doc_updated["vorgangsID"]

        # should not work without specifying
        # "treat_list_as_element = True" in nested_update
        self.assertEqual(plz1, "82269abcdef")
        self.assertEqual(plz2, "86899abcdef")
        self.assertEqual(vorgangsid, "1abcdef")

    def test_nested_alter_list_input_with_args_in_place_false(self):
        # callback functions
        def callback(data, str1, str2):
            return str(data) + str1 + str2

        doc_updated = nested_alter(
            self.sample_data4,
            ["plz", "vorgangsID"],
            callback,
            function_parameters=["abc", "def"],
            in_place=False,
        )

        plz1 = doc_updated["plz"]
        plz2 = doc_updated["vertragsteile"][0]["beitragsDaten"]["plz"]
        vorgangsid = doc_updated["vorgangsID"]

        # should not work without specifying
        # "treat_list_as_element = True" in nested_update
        self.assertEqual(plz1, "82269abcdef")
        self.assertEqual(plz2, "86899abcdef")
        self.assertEqual(vorgangsid, "1abcdef")

    def test_nested_alter_list_input_in_place_false(self):
        # callback functions
        def callback(data):
            return str(data) + "###"

        doc_updated = nested_alter(
            self.sample_data4, ["plz", "vorgangsID"], callback, in_place=False
        )

        plz1 = doc_updated["plz"]
        plz2 = doc_updated["vertragsteile"][0]["beitragsDaten"]["plz"]
        vorgangsid = doc_updated["vorgangsID"]

        # should not work without specifying
        # "treat_list_as_element = True" in nested_update
        self.assertEqual(plz1, "82269###")
        self.assertEqual(plz2, "86899###")
        self.assertEqual(vorgangsid, "1###")

    def test_nested_alter_taco_for_example(self):
        documents = [{"taco": 42}, {"salsa": [{"burrito": {"taco": 69}}]}]

        # write a callback function which processes a scalar value.
        # Be aware about the possible types which can be passed to
        # the callback functions.
        # In this example we can be sure that only int will be passed,
        # in production you should check the type because it could be
        # anything.
        def callback(data):
            return data + 10  # (data/100*10)

        # The alter-version only works for scalar input (one dict),
        # if you need to adress a list of dicts, you have to
        # manually iterate over those and pass them to
        # nested_update one by one
        out = []
        for elem in documents:
            altered_document = nested_alter(elem, "taco", callback)
            out.append(altered_document)

        self.maxDiff = None
        self.assertEqual(out[0]["taco"], 52)
        self.assertEqual(out[1]["salsa"][0]["burrito"]["taco"], 79)

    def test_nested_alter_work_with_right_order(self):
        document = {"taco": 42, "salsa": [{"burrito": {"key": 20}}], "key": 50}

        def callback(data):
            return data + 100

        altered_document = nested_alter(document, "key", callback, in_place=True)

        self.assertEqual(altered_document["salsa"][0]["burrito"]["key"], 120)
        self.assertEqual(altered_document["key"], 150)

    def test_sample_data4(self):
        result = {
            "modelversion": "1.1.0",
            "vorgangsID": "1",
            "versorgungsvorschlagDatum": 1510558834978,
            "eingangsdatum": 1510558834978,
            "plz": 82270,
            "vertragsteile": [
                {
                    "typ": "1",
                    "beitragsDaten": {
                        "endalter": 85,
                        "brutto": 58.76,
                        "netto": 58.76,
                        "zahlungsrhythmus": "MONATLICH",
                        "plz": 86900,
                    },
                    "beginn": 1512082800000,
                    "lebenslang": "True",
                    "ueberschussverwendung": {
                        "ueberschussverwendung": "2",
                        "indexoption": "3",
                    },
                    "deckung": [
                        {
                            "typ": "2",
                            "art": "1",
                            "leistung": {"value": 7500242424.0, "einheit": "2"},
                            "leistungsRhythmus": "1",
                        }
                    ],
                    "zuschlagNachlass": [],
                },
                {
                    "typ": "1",
                    "beitragsDaten": {
                        "endalter": 85,
                        "brutto": 0.6,
                        "netto": 0.6,
                        "zahlungsrhythmus": "1",
                    },
                    "zuschlagNachlass": [],
                },
            ],
        }

        # add +1 to all plz
        def callback(data):
            return data + 1

        self.maxDiff = None
        self.assertEqual(result, nested_alter(self.sample_data4, "plz", callback))



import os
from pathlib import Path
from typing import Type

from config_file.parsers.abstract_parser import AbstractParser
from config_file.parsers.ini_parser import IniParser
from config_file.parsers.json_parser import JsonParser
from config_file.parsers.toml_parser import TomlParser
from config_file.parsers.yaml_parser import YamlParser

from .utils import split_on_dot


class ConfigFilePath(type(Path())):
    """A ConfigFilePath object, which subclasses Path.

    It provides us easy ways to grab the original config file
    path, file extension, validate the path, or determine the parser
    that should be use for the given file type.
    """

    @property
    def original_path(self) -> "ConfigFilePath":
        """Computes the original configuration file path.

        e.g. ConfigFilePath("./config.ini").original_path -> "./config.original.ini"

        Returns:
            The path to the original configuration file, which
            may or may not exist.
        """
        file_parts = split_on_dot(self, only_last_dot=True)
        file_parts.insert(-1, "original")
        return ConfigFilePath(".".join(file_parts))

    @property
    def extension(self) -> str:
        """Retrieves the extension of the file
        path we passed in, without the '.'.

        e.g. ConfigFilePath("./pyproject.toml").extension -> "toml"

        Returns:
            The extension of the current file path.
        """
        _, extension = os.path.splitext(str(self))
        return extension[1:]

    @property
    def parser(self) -> Type[AbstractParser]:
        """Determine what parser should be used for this file.

        Raises:
            ValueError: If the extension of the file is not recognized.
            ParsingError: If the contents could not be parsed.

        Returns:
            The instantiated parser that should be used for this file.
        """
        if self.extension == "":
            raise ValueError(
                "Tried to determine a parser to use, but the file at "
                f"{self} does not have an extension."
            )

        if self.extension == "ini":
            return IniParser(self.contents)
        elif self.extension == "json":
            return JsonParser(self.contents)
        elif self.extension == "yaml" or self.extension == "yml":
            return YamlParser(self.contents)
        elif self.extension == "toml":
            return TomlParser(self.contents)
        else:
            raise ValueError(
                f"File path at `{self}` contains an unrecognized file type."
            )

    @property
    def contents(self) -> str:
        """Retrieve the contents of the file.

        Returns:
            The contents of the file at this path.
        """
        with open(self, "r") as file:
            return file.read()

    def validate(self) -> "ConfigFilePath":
        """Validates that the current path exists, is a directory,
        and expands any home tidles.

        Raises:
            FileNotFoundError: If the path does not exist.
            ValueError: If the path leads to a directory.

        Returns:
            A ConfigFilePath object with any ~'s expanded.
        """
        if len(self.parts) >= 1 and self.parts[0] == "~":
            self = ConfigFilePath(self.expanduser())

        if not self.exists():
            raise FileNotFoundError(
                f"The specified config file ({self}) does not exist."
            )

        if self.is_dir():
            raise ValueError(f"The specified config file ({self}) is a directory")

        return self



from datetime import datetime

import pytest

from config_file.parsers.parse_value import (
    can_be_parsed_as_bool,
    can_be_parsed_as_dict,
    can_be_parsed_as_float,
    can_be_parsed_as_int,
    can_be_parsed_as_list,
    parse_value,
)


@pytest.mark.parametrize(
    "test_input,expected",
    [
        ("5", 5),
        ("-5", -5),
        ("0", 0),
        ("-234.5", -234.5),
        ("35.5", 35.5),
        ("TRUE", True),
        ("true", True),
        ("false", False),
    ],
)
def test_parse_value_parses_ints_floats_and_bools(test_input, expected):
    """
    config_file.parse_value.parse_value

    Parses string values that contain ints, floats, and booleans into
    their native representation.
    """
    assert parse_value(test_input) == expected


@pytest.mark.parametrize(
    "test_input,expected",
    [
        ({"blah": "5"}, {"blah": 5}),
        (
            {"blah": {"blah2": {"blah3": "fAlSe"}, "abc": "5.25", "boo": "tRuE"}},
            {"blah": {"blah2": {"blah3": False}, "abc": 5.25, "boo": True}},
        ),
        ("{ 'key': '5', 'key2': 'fAlSE' }", {"key": 5, "key2": False}),
        (
            ["5", "4.4", "faLSe", "trUE", "foo", "-5", "-5.2"],
            [5, 4.4, False, True, "foo", -5, -5.2],
        ),
        (" [1, 2, 3] ", [1, 2, 3]),
        ("  [True, [1, 5], {}, 5.5]", [True, [1, 5], {}, 5.5]),
        ("  [ [ [ [] ]] ]  ", [[[[]]]]),
    ],
)
def test_parse_value_parses_dicts_and_lists(test_input, expected):
    """
    config_file.parse_value.parse_value

    Parses string values that contain dictionaries or lists into
    their native representation.
    """
    assert parse_value(test_input) == expected


@pytest.mark.parametrize("test_input", ["blah", datetime(2020, 1, 1)])
def test_parse_value_returns_values_that_cannot_be_parsed(test_input):
    """
    config_file.parse_value.parse_value

    If a value is given that cannot be parsed by parse_value,
    it should simply return that value back.
    """
    assert parse_value(test_input) == test_input


@pytest.mark.parametrize(
    "test_input",
    [(343, 343), (3.2, 3.2), (True, True), (False, False), ({}, {}), ([], [])],
)
def test_parse_value_returns_values_that_are_already_in_their_native_type(test_input):
    """
    config_file.parse_value.parse_value

    Returns back the inputted value if it is already in its native type.
    """
    assert parse_value(test_input) == test_input


@pytest.mark.parametrize("value", [-1, 0, 1, "1", "-1", "   0   "])
def test_values_that_can_be_parsed_as_int(value):
    """
    config_file.parse_value.can_be_parsed_as_int

    Returns True for values that can be parsed as an int.
    """
    assert can_be_parsed_as_int(value)


@pytest.mark.parametrize(
    "value", [True, False, 5.5, "5.5", [], {}, "ruff", datetime.now()]
)
def test_values_that_cannot_be_parsed_as_int(value):
    """
    config_file.parse_value.can_be_parsed_as_int

    Returns False for values that cannot be parsed as an int.
    """
    assert not can_be_parsed_as_int(value)


@pytest.mark.parametrize("value", [5.5, -1.2, ".1", "5.5", "   1.1  ", "-0.0", "0.0"])
def test_values_that_can_be_parsed_as_float(value):
    """
    config_file.parse_value.can_be_parsed_as_float

    Returns True for values that can be parsed as an float.
    """
    assert can_be_parsed_as_float(value)


@pytest.mark.parametrize(
    "value",
    [
        545,
        0,
        -3,
        "1",
        "-1",
        True,
        False,
        "   0   ",
        "0..0",
        [],
        {},
        "ruff",
        datetime.now(),
    ],
)
def test_values_that_cannot_be_parsed_as_float(value):
    """
    config_file.parse_value.can_be_parsed_as_float

    Returns False for values that cannot be parsed as an float.
    """
    assert not can_be_parsed_as_float(value)


@pytest.mark.parametrize(
    "value", [True, False, "false", "true", "TRUE", "FALSE", "  FALSE  ", "  TRUE  "]
)
def test_values_that_can_be_parsed_as_bool(value):
    """
    config_file.parse_value.can_be_parsed_as_bool

    Returns True for values that can be parsed as a bool.
    """
    assert can_be_parsed_as_bool(value)


@pytest.mark.parametrize("value", [0, 1, -1, 5.3, "ruff", {}, [], datetime.now()])
def test_values_that_cannot_be_parsed_as_bool(value):
    """
    config_file.parse_value.can_be_parsed_as_bool

    Returns False for values that cannot be parsed as a bool.
    """
    assert not can_be_parsed_as_bool(value)


@pytest.mark.parametrize("value", [{}, {"hello": 5}, "{}", "  {  }  ", "{'test': 5}"])
def test_values_that_can_be_parsed_as_dict(value):
    """
    config_file.parse_value.can_be_parsed_as_dict

    Returns True for values that can be parsed as a dict.
    """
    assert can_be_parsed_as_dict(value)


@pytest.mark.parametrize("value", [0, 5, -1, 5.5, [], "ruff", " { invalid } "])
def test_values_that_cannot_be_parsed_as_dict(value):
    """
    config_file.parse_value.can_be_parsed_as_dict

    Returns False for values that cannot be parsed as a dict.
    """
    assert not can_be_parsed_as_dict(value)


@pytest.mark.parametrize("value", [[], [1, 2, True, 5.5], "[]", " [] ", " [1, 2, 3]"])
def test_values_that_can_be_parsed_as_list(value):
    """
    config_file.parse_value.can_be_parsed_as_list

    Returns True for values that can be parsed as a list.
    """
    assert can_be_parsed_as_list(value)


@pytest.mark.parametrize("value", [0, 5, -1, 5.5, {}, "ruff", " [ invalid ] "])
def test_values_that_cannot_be_parsed_as_list(value):
    """
    config_file.parse_value.can_be_parsed_as_list

    Returns False for values that cannot be parsed as a list.
    """
    assert not can_be_parsed_as_list(value)



import configparser
from io import StringIO
from typing import Any, Type

from .base_parser import BaseParser


class IniParser(BaseParser):
    def __init__(self, file_contents: str):
        super().__init__(file_contents)

    @property
    def decode_error(self) -> Type[Exception]:
        return configparser.Error

    def loads(self, contents: str) -> dict:
        parser = configparser.ConfigParser()
        parser.read_string(contents)

        return self.__create_configparser_dict(parser)

    def dumps(self, loaded_contents: dict) -> str:
        parser = configparser.ConfigParser()
        buffer = StringIO()

        parser.read_dict(loaded_contents)
        parser.write(buffer)

        return buffer.getvalue()

    def __create_configparser_dict(self, parser: configparser.ConfigParser) -> dict:
        result = {}
        items = dict(parser.items())

        for section in parser.sections():
            result.update({str(section): dict(items[section])})

        return result

    def get(self, search_key: str) -> Any:
        retrieved_value = super().get(search_key)
        return (
            retrieved_value if type(retrieved_value) is dict else str(retrieved_value)
        )



from pathlib import Path

import pytest

from config_file.config_file_path import ConfigFilePath
from config_file.parsers.ini_parser import IniParser
from config_file.parsers.json_parser import JsonParser
from config_file.parsers.toml_parser import TomlParser
from config_file.parsers.yaml_parser import YamlParser


def test_config_file_path_is_a_path():
    """
    config_file.config_file_path.ConfigFilePath

    Ensure that the ConfigFilePath object is an instance
    of Path from pathlib.
    """
    assert isinstance(ConfigFilePath("."), Path)


def test_config_file_path_is_able_to_read_the_contents(template_file):
    """
    config_file.config_file_path.ConfigFilePath

    Ensure that the contents retrieved are the same
    as the contents of the file.
    """
    template = template_file("ini", "default")
    path = ConfigFilePath(template)

    assert path.contents == template.read_text()


def test_config_file_path_is_able_to_calculate_the_original_path():
    """
    config_file.config_file_path.ConfigFilePath

    Ensure that the calculated original configuration
    path includes a "original" between the filename and
    extension.
    """
    path = ConfigFilePath("path-existing-does-not-matter.txt")

    assert str(path.original_path) == "path-existing-does-not-matter.original.txt"


def test_config_file_path_is_able_to_retrieve_the_extension():
    """
    config_file.config_file_path.ConfigFilePath

    Ensure that the retrieved extension is correct
    for the given filepath and that it does not retrieve
    the dot (.) from the extension.
    """
    path = ConfigFilePath("some-file.txt")

    assert path.extension == "txt"


@pytest.mark.parametrize(
    "template_file_type, expected_parser_type",
    [
        ("ini", IniParser),
        ("json", JsonParser),
        ("yaml", YamlParser),
        ("toml", TomlParser),
    ],
)
def test_config_file_path_decides_on_the_correct_parser(
    template_file, template_file_type, expected_parser_type
):
    """
    config_file.config_file_path.ConfigFilePath

    Ensure that a filepath with an extension that is supported
    retrieves the correct parser for it.
    """
    template = template_file(template_file_type, "default")
    path = ConfigFilePath(template)

    assert isinstance(path.parser, expected_parser_type)


@pytest.mark.parametrize("filename", ["some-file", "some-file.unsupported-extension"])
def test_config_file_path_raises_value_error_on_parser_if_invalid_extension(filename):
    """
    config_file.config_file_path.ConfigFilePath

    Ensure that a ValueError is raised if we have a
    filepath with no extension or a filepath with an
    unsupported extension.
    """
    path = ConfigFilePath(filename)

    with pytest.raises(ValueError):
        path.parser



from pathlib import Path
from shutil import copyfile
from typing import Any, Union

from config_file.config_file_path import ConfigFilePath
from config_file.parsers.parse_value import parse_value
from config_file.utils import Default


class ConfigFile:
    def __init__(self, file_path: Union[str, Path]) -> None:
        """
        Stores the config file path and expands it if needed, reads in
        the file contents, and determines what parser should be used for
        the given file.

        Args:
            file_path: The path to your configuration file.

        Raises:
            ValueError: If the specified file path does not have an extension
                        that is supported or it is a directory.
            FileNotFoundError: If the specified file path does not exist.
            ParsingError: If the file_path could not be parsed.
        """
        self.__path = ConfigFilePath(file_path).validate()
        self.__parser = self.__path.parser

    @property
    def path(self) -> Path:
        return Path(self.__path)

    @property
    def original_path(self) -> Path:
        return Path(self.__path.original_path)

    def __getitem__(self, key: str) -> Any:
        return self.__parser.parsed_content[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.__parser.parsed_content[key] = value

    def __delitem__(self, key: str) -> None:
        del self.__parser.parsed_content[key]

    def __contains__(self, key: str) -> bool:
        return self.has(key)

    def __str__(self) -> str:
        return str(self.__parser)

    def __repr__(self) -> str:
        return f"{self.path}\n\n{str(self.__parser)}"

    def get(
        self,
        key: str,
        parse_types: bool = False,
        return_type: Any = None,
        default: Any = Default(None),
    ) -> Any:
        """
        Retrieve the value of a key.

        Args:
            key: The key to retrieve.
            parse_types: Automatically parse ints, floats, booleans, dicts, and
                         lists. This recursively parses all types in whatever you're
                         retrieving, not just a single type.

                         e.g. If you are retrieving a section, all values in that
                         section will be parsed.

            return_type: The type to coerce the return value to.
            default: The default value to return if the value of the key is empty.

        Returns:
            The value of the key.

        Raises:
            KeyError: If a key is attempted to be retrieved that does not exist.
            ValueError: If the value is not able to be coerced into return_type.
        """
        try:
            key_value = self.__parser.get(key)
        except KeyError as error:
            if isinstance(default, Default) and default.value is None:
                raise error
            else:
                key_value = default

        if parse_types:
            key_value = parse_value(key_value)

        return return_type(key_value) if return_type else key_value

    def set(self, key: str, value: Any) -> None:
        """Sets the value of a key.

        If the given key does not exist, it will be automatically
        created for you. That includes if there are multiple keys
        in a row that do not exist.
        e.g. set('exists.does_not.does_not.also_does_not', 5)

        The behavior of how this is done, however, depends on the
        file used. For example, with INI, subsections are not supported.
        So it would create a key in the section `exists` called `does_not`
        and set it to the value {'does_not': {'also_does_not': 5}}.

        Args:
            key: The section, sub-section, or key to delete.
            value: The value to set the key to.
        """
        self.__parser.set(key, value)

    def delete(self, key: str) -> None:
        """Deletes a section or key.

        Args:
            key: The section, sub-section, or key to delete.

        Raises:
            KeyError: If a key is attempted to be deleted that
            does not exist.
        """
        self.__parser.delete(key)

    def has(self, key: str, wild: bool = False) -> bool:
        """
        Check if a section, sub-section, or key exists.

        Some formats, like JSON, do not have sections and
        therefore, it would only be checking if a particular
        key exists.

        Args:
            key: The section, sub-section, or key to find.
            wild: Whether or not to search everywhere.

            Without `wild`, a single word `key` without a `.`
            will look at the outer most hierarchy of the file for it.

            With `wild`, that single word `key` will be searched
            for throughout the entire file.

        Returns:
            True if the key exists. False otherwise.
        """
        return self.__parser.has(key, wild=wild)

    def restore_original(self, original_path: Union[str, Path, None] = None) -> None:
        """Restores the original the configuration file.

        The current one is deleted and the original is copied back
        in its place. The internal contents are then reset to the
        new file.

        Args:
            original_path: The original file to reset to.

            Defaults to the original_path property if it is not
            provided.

        Raises:
            FileNotFoundError: If the original configuration file (whether
            calculated or passed in) does not exist or if the current
            configuration path is passed in as the original_path (since
            it is deleted before the original file is copied over).

            OSError: If the current configuration path is not writable.
        """
        original_path = ConfigFilePath(
            original_path if original_path else self.original_path
        ).validate()

        self.__path.unlink()
        copyfile(original_path, self.__path)
        self.__parser.reset_internal_contents(self.__path.contents)

    def save(self) -> None:
        """
        Save your configuration changes.

        This writes the file back out, including any changes
        you've made, to the specified path given from this
        object's constructor.
        """
        with open(str(self.__path), "w") as config_file:
            config_file.write(str(self))
